
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "php.h"
#include "php_ini.h"
#include "ext/standard/info.h"
#include "php_clips.h"
#include "clipssrc/clips.h"

#define VERSION "0.3.1"



/* Function Entry - Register functions in PHP */
function_entry clips_functions[] = {
	PHP_FE(clips_load, NULL)
	PHP_FE(clips_binary_load, NULL)
	PHP_FE(clips_reset, NULL)
	PHP_FE(clips_clear, NULL)
	PHP_FE(clips_run, NULL)
	PHP_FE(clips_get_fact_list, NULL)
	PHP_FE(clips_get_instance_list, NULL)
	PHP_FE(clips_set_strategy, NULL)
	{NULL, NULL, NULL}
};



/* Module Entry - Register the extension in PHP */
zend_module_entry clips_module_entry = {
	STANDARD_MODULE_HEADER,
	"clips",
	clips_functions,
	PHP_MINIT(clips),
	NULL,
	PHP_RINIT(clips),
	NULL,
	PHP_MINFO(clips),
	VERSION,
	STANDARD_MODULE_PROPERTIES
};

#ifdef COMPILE_DL_CLIPS
ZEND_GET_MODULE(clips)
#endif



/* Module Initialization Function -
   Is called when the webserver starts */
PHP_MINIT_FUNCTION(clips) {

	/* Register CLIPS Constants */
	REGISTER_LONG_CONSTANT("DEPTH_STRATEGY", DEPTH_STRATEGY, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("BREADTH_STRATEGY", BREADTH_STRATEGY, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("LEX_STRATEGY", LEX_STRATEGY, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("MEA_STRATEGY", MEA_STRATEGY, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("COMPLEXITY_STRATEGY", COMPLEXITY_STRATEGY, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("SIMPLICITY_STRATEGY", SIMPLICITY_STRATEGY, CONST_CS | CONST_PERSISTENT);
	REGISTER_LONG_CONSTANT("RANDOM_STRATEGY", RANDOM_STRATEGY, CONST_CS | CONST_PERSISTENT);

	return SUCCESS;
}



/* Request Initialization Function -
   Is called upon each page request */
PHP_RINIT_FUNCTION(clips) {

	/* Initialize a CLIPS environment */
	InitializeEnvironment();

	return SUCCESS;
}



/* Module Info Function -
   Defines information about the extension to be listed when 
   phpinfo() is called */
PHP_MINFO_FUNCTION(clips) {
	php_info_print_table_start();
	php_info_print_table_row(2, "CLIPS Support", "enabled");
	php_info_print_table_row(2, "CLIPS Extension Version", VERSION);
	php_info_print_table_row(2, "CLIPS Version", "6.2");
	php_info_print_table_end();
}






/* Implementation */



/* {{{ proto bool clips_load(string file); 
   Load a set of constructs into CLIPS from a file */
PHP_FUNCTION(clips_load) {
	char *file;
	int file_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &file, &file_len) == FAILURE) {
        RETURN_FALSE;
    }

	if (Load(file)) {
		RETURN_TRUE;
	} else {
		zend_error(E_WARNING, "failed to load clips source file '%s'", file);
		RETURN_FALSE;
	}
}
/* }}} */



/* {{{ proto bool clips_binary_load(string file); 
   Load a set of constructs into CLIPS from a binary CLIPS file */
PHP_FUNCTION(clips_binary_load) {
	char *file;
	int file_len;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s", &file, &file_len) == FAILURE) {
        RETURN_FALSE;
    }

	if (Bload(file)) {
		RETURN_TRUE;
	} else {
		zend_error(E_WARNING, "failed to load clips binary file '%s'", file);
		RETURN_FALSE;
	}
}
/* }}} */



/* {{{ proto void clips_reset();
   Reset the CLIPS environment */
PHP_FUNCTION(clips_reset) {
	Reset();
	RETURN_TRUE;
}
/* }}} */



/* {{{ proto void clips_clear();
   Clear the CLIPS environment */
PHP_FUNCTION(clips_clear) {
	Clear();
	RETURN_TRUE;
}
/* }}} */



/* {{{ proto int clips_run([int run_limit]);
   Execute rules in the agenda, returns the number of rules that were fired. */
PHP_FUNCTION(clips_run) {
	int run_limit = -1;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|l", &run_limit) == FAILURE) {
        RETURN_FALSE;
    }

	RETURN_LONG(Run(run_limit));
}
/* }}} */



/* {{{ proto array clips_get_fact_list([string module]);
   Retrieve the fact list */
PHP_FUNCTION(clips_get_fact_list) {
	void *module_ptr, *multifield_ptr, *multifield_ptr2, *fact_ptr;
	char *module;
	int module_len, i, end, i2, end2;
	zval *cur_fact;
	DATA_OBJECT fact_list, slot_names, slot_value;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "|s", &module, &module_len) == FAILURE) {
        RETURN_FALSE;
    }

	if (ZEND_NUM_ARGS() == 0) {
		GetFactList(&fact_list, NULL);
	} 
	
	else {
		/* If a module name was given, get a pointer to the module */
		module_ptr = FindDefmodule(module);

		if (module_ptr == NULL) {
			zend_error(E_WARNING, "Undefined module '%s'", module);
			RETURN_FALSE;
		}

		GetFactList(&fact_list, module_ptr);
	}

	if (GetType(fact_list) != MULTIFIELD) {
		RETURN_FALSE;
	}


	/* Initialize the return value to an empty array */
	array_init(return_value);
	
	/* Find the end/length of the fact list */
	end = GetDOEnd(fact_list);

	/* Get the fact list multifield pointer */
	multifield_ptr = GetValue(fact_list);

	/* Iterate through the list of fact pointers */
	for (i = GetDOBegin(fact_list); i <= end; i++) {

		/* Do this only if the current item is a fact pointer */
		if (GetMFType(multifield_ptr, i) == FACT_ADDRESS) {

			/* Retrieve the fact pointer */
			fact_ptr = GetMFValue(multifield_ptr, i);

			/* Get the slot names of the current fact */
			FactSlotNames(fact_ptr, &slot_names);

			/* Initialize an array to contain an array representing the current fact */
			MAKE_STD_ZVAL(cur_fact);
			array_init(cur_fact);
			
			end2 = GetDOEnd(slot_names);
			multifield_ptr2 = GetValue(slot_names);

			/* Iterate through the fact slots and retrieve the values in each slot */
			for (i2 = GetDOBegin(slot_names); i2 <= end2; i2++) {
				
				/* Store the current slot value into a new DATA_OBJECT */
				GetFactSlot(fact_ptr, ValueToString(GetMFValue(multifield_ptr2, i2)), &slot_value);
				
				/* Convert the DATA_OBJECT into a PHP data type and append to the fact array */
				switch (GetType(slot_value)) {
					case FLOAT: 
						add_assoc_double(cur_fact, ValueToString(GetMFValue(multifield_ptr2, i2)), DOToDouble(slot_value));
						break;
					
					case STRING: case SYMBOL: case INSTANCE_NAME: 
						add_assoc_string(cur_fact, ValueToString(GetMFValue(multifield_ptr2, i2)), DOToString(slot_value), 1);
						break;
					
					case INTEGER: 
						add_assoc_long(cur_fact, ValueToString(GetMFValue(multifield_ptr2, i2)), DOToLong(slot_value));
						break;
				}
			}

			/* Add the fact array to the return value array */
			add_next_index_zval(return_value, cur_fact);
		}
	}
}
/* }}} */



/* {{{ proto array clips_get_instance_list();
   Retrieve the instance list */
PHP_FUNCTION(clips_get_instance_list) {
	DATA_OBJECT instance_list;
	DATA_OBJECT slot_names;
	DATA_OBJECT slot_value;

	void *multifield_ptr;
	void *instance_ptr;
	void *defclass_ptr;
	int i, end;
	zval *cur_instance;

	/* Initialize the return value to an empty array */
	array_init(return_value);
	
	/* Find the end/length of the instance list */
	end = GetDOEnd(instance_list);

	/* Get the instance list multifield pointer */
	multifield_ptr = GetValue(instance_list);

	/* Get a pointer to the first instance */
	instance_ptr = GetNextInstance(instance_ptr);

	/* Iterate through the list of instance pointers */
	do {

		/* Get a pointer to the instance's defclass */
		defclass_ptr = GetInstanceClass(instance_ptr);

		/* Get the slot names of the current instance */
		ClassSlots(defclass_ptr, &slot_names, 1);

		/* Initialize an array to contain an array representing the current instance */
		MAKE_STD_ZVAL(cur_instance);
		array_init(cur_instance);
		
		end = GetDOEnd(slot_names);
		multifield_ptr = GetValue(slot_names);

		/* Iterate through the instance slots and retrieve the values in each slot */
		for (i = GetDOBegin(slot_names); i <= end; i++) {
			
			/* Store the current slot value into a new DATA_OBJECT */
			DirectGetSlot(instance_ptr, ValueToString(GetMFValue(multifield_ptr, i)), &slot_value);
			
			/* Convert the DATA_OBJECT into a PHP data type and append to the instance array */
			switch (GetType(slot_value)) {
				case FLOAT: 
					add_assoc_double(cur_instance, ValueToString(GetMFValue(multifield_ptr, i)), DOToDouble(slot_value));
					break;
				
				case STRING: case SYMBOL: case INSTANCE_NAME: 
					add_assoc_string(cur_instance, ValueToString(GetMFValue(multifield_ptr, i)), DOToString(slot_value), 1);
					break;
				
				case INTEGER: 
					add_assoc_long(cur_instance, ValueToString(GetMFValue(multifield_ptr, i)), DOToLong(slot_value));
					break;
			}
		}

		/* Add the instance array to the return value array */
		add_next_index_zval(return_value, cur_instance);

		/* Get a pointer to the next instance */
		instance_ptr = GetNextInstance(instance_ptr);
	} while (instance_ptr != NULL);
}
/* }}} */



/* {{{ proto void set_strategy(int value);
   Set the conflict resolution strategy */
PHP_FUNCTION(clips_set_strategy) {
	int strategy;

	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "l", &strategy) == FAILURE) {
        RETURN_FALSE;
    }

	SetStrategy(strategy);
}
/* }}} */

